from pathlib import Path
from typing import Any

from urlpath import URL

from yaml_ld.document_loaders import content_types
from yaml_ld.document_loaders.base import (
    DocumentLoader,
    DocumentLoaderOptions,
    PyLDResponse,
)
from yaml_ld.errors import LoadingDocumentFailed


class HTTPDocumentLoader(DocumentLoader):
    """Load documents from HTTP sources."""

    def __call__(
        self,
        source: str | Path,
        options: DocumentLoaderOptions,
    ) -> PyLDResponse:
        """Load documents from HTTP sources."""
        url = URL(source)

        raw_content = url.get(stream=True).raw
        raw_content.decode_content = True

        content_type = content_types.by_extension(url.suffix)
        if content_type is None:
            raise ValueError(f'What content type is extension {url.suffix}?')

        parser = content_types.parser_by_content_type(content_type)
        if parser is None:
            raise LoadingDocumentFailed(path=source)

        yaml_document = parser(raw_content, str(source), options)

        return {
            'document': yaml_document,
            'documentUrl': str(source),
            'contextUrl': None,
            'contentType': 'application/ld+yaml',
        }



import functools
from dataclasses import dataclass

from documented import DocumentedError

from yaml_ld.document_parsers.base import BaseDocumentParser
from yaml_ld.document_parsers.yaml_parser import YAMLDocumentParser


def by_extension(extension: str) -> str | None:
    return {
        '.json': 'application/json',
        '.jsonld': 'application/ld+json',
        '.jldt': 'application/ld+json',
        '.jldte': 'application/ld+json',
        '.yaml': 'application/yaml',
        '.yamlld': 'application/ld+yaml',
        '.html': 'text/html',
    }.get(extension)


@functools.cache
def parser_by_content_type_map():
    """
    Map content types to parsers.

    FIXME: Make this dynamic; perhaps a plugin entry point even.
    """
    # This prevents a cyclic import problem.
    from yaml_ld.document_parsers.html_parser import (  # noqa: WPS433
        HTMLDocumentParser,
    )

    return {
        'application/json': YAMLDocumentParser,
        'application/ld+json': YAMLDocumentParser,
        'application/yaml': YAMLDocumentParser,
        'application/ld+yaml': YAMLDocumentParser,
        'text/html': HTMLDocumentParser,
    }


@dataclass
class ParserNotFound(DocumentedError):   # type: ignore
    """Cannot find a parser for `{self.content_type}` content type."""

    content_type: str


def parser_by_content_type(content_type: str) -> BaseDocumentParser:
    """Find a parser based on content type."""
    try:
        return parser_by_content_type_map()[content_type]()
    except KeyError:
        raise ParserNotFound(content_type)



import io
from dataclasses import dataclass
from typing import Iterable

import lxml  # noqa: S410
from pyld.jsonld import JsonLdError, parse_url, prepend_base

from yaml_ld.document_loaders.content_types import (
    ParserNotFound,
    parser_by_content_type,
)
from yaml_ld.document_parsers.base import (
    BaseDocumentParser,
    DocumentLoaderOptions,
)
from yaml_ld.errors import DocumentIsScalar, NoLinkedDataFoundInHTML
from yaml_ld.models import JsonLdRecord


@dataclass
class Script:
    """HTML <script> tag."""

    content_type: str
    content: str   # noqa: WPS110


class HTMLDocumentParser(BaseDocumentParser):
    """Parse HTML documents, specifically their <script> tags."""

    def __call__(
        self,
        data_stream: io.TextIOBase,
        source: str,
        options: DocumentLoaderOptions,
    ) -> JsonLdRecord | list[JsonLdRecord]:
        """Parse HTML with LD in <script> tags."""
        scripts = self.extract_script_tags(
            html_content=data_stream.read(),
            url=source,
            profile=None,
            options=options,
        )

        documents = self.parsed_documents_stream(
            scripts=scripts,
            source=source,
            options=options,
        )

        if options.get('extractAllScripts'):
            return list(documents)

        try:
            return next(iter(documents))
        except StopIteration:
            raise NoLinkedDataFoundInHTML()

    def extract_script_tags(   # noqa: C901, WPS210
        self,
        html_content: str,
        url,
        profile,
        options,
    ) -> Iterable[Script]:
        """Load one or more script tags from an HTML source."""
        document = lxml.html.fromstring(html_content)
        # potentially update options[:base]
        html_base = document.xpath('/html/head/base/@href')
        if html_base:
            # use either specified base, or document location
            effective_base = options.get('base', url)
            if effective_base:
                html_base = prepend_base(effective_base, html_base[0])
            options['base'] = html_base

        url_elements = parse_url(url)
        if url_elements.fragment:
            # FIXME: CGI decode
            fragment_id = url_elements.fragment
            singular_element = document.xpath(
                '//script[@id="%s"]' % fragment_id,
            )
            if not singular_element:
                raise JsonLdError(
                    'No script tag found for id.',
                    'jsonld.LoadDocumentError',
                    {'id': fragment_id},
                    code='loading document failed',
                )

            yield Script(
                content_type=singular_element[0].xpath('@type')[0],
                content=singular_element[0].text_content(),
            )

        elements = document.xpath('//script')
        for element in elements:   # noqa: WPS526
            yield Script(
                content_type=element.xpath('@type')[0],
                content=element.text_content(),
            )

    def parsed_documents_stream(
        self,
        scripts: Iterable[Script],
        source: str,
        options: DocumentLoaderOptions,
    ) -> Iterable[JsonLdRecord]:
        """Parse each of the given scripts and emit a stream of LD documents."""
        for script in scripts:
            try:
                parser = parser_by_content_type(
                    script.content_type,
                )
            except ParserNotFound:
                continue

            stream = io.StringIO(script.content)
            document_or_array = parser(stream, source, options)

            match document_or_array:
                case list() as array:
                    yield from array

                case dict() as mapping:
                    yield mapping

                case scalar:
                    raise DocumentIsScalar(scalar)



from abc import ABC, abstractmethod
from pathlib import Path

from typing_extensions import TypedDict

from yaml_ld.models import JsonLdRecord

PyLDResponse = TypedDict(
    'PyLDResponse', {
        'contentType': str,
        'contextUrl': str | None,   # noqa: WPS465
        'documentUrl': str,
        'document': JsonLdRecord | list[JsonLdRecord],
    },
)

DocumentLoaderOptions = TypedDict(
    'DocumentLoaderOptions',
    {
        'extractAllScripts': bool,
    },
)


class DocumentLoader(ABC):
    @abstractmethod
    def __call__(
        self,
        source: str | Path,
        options: DocumentLoaderOptions,
    ) -> PyLDResponse:
        raise NotImplementedError()



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import BaseOptions, ExtractAllScriptsOptions, JsonLdInput
from yaml_ld.rdf import Dataset


class ToRDFOptions(BaseOptions, ExtractAllScriptsOptions):   # type: ignore
    """Options for converting YAML-LD to RDF."""

    format: str = 'application/n-quads'
    """The format to use to output a string: 'application/n-quads'
    for N-Quads."""

    produce_generalized_rdf: bool = False
    """True to output generalized RDF, false to produce only standard RDF."""

    rdf_direction: str = 'i18n-datatype'
    """Only 'i18n-datatype' supported."""


@validate_call(config=dict(arbitrary_types_allowed=True))
def to_rdf(
    document: JsonLdInput,
    options: ToRDFOptions = ToRDFOptions(),  # type: ignore
) -> Dataset:
    """Convert a YAML-LD document to RDF."""
    with except_json_ld_errors():
        return jsonld.to_rdf(
            document,
            options=options.model_dump(
                by_alias=True,
                exclude_defaults=True,
            ),
        )



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.models import BaseOptions


@validate_call(config=dict(arbitrary_types_allowed=True))
def load_document(
    url,
    base=None,
    profile=None,
    requestProfile=None,
    options: BaseOptions = BaseOptions(),   # type: ignore
):
    return jsonld.load_document(
        url=url,
        options=options.model_dump(
            by_alias=True,
            exclude_defaults=True,
        ),
        base=base,
        profile=profile,
        requestProfile=requestProfile,
    )



import io
from abc import abstractmethod

from yaml_ld.document_loaders.base import DocumentLoaderOptions
from yaml_ld.models import JsonLdRecord


class BaseDocumentParser:
    """Parse documents of various types into LD."""

    @abstractmethod
    def __call__(
        self,
        data_stream: io.TextIOBase,
        source: str,
        options: DocumentLoaderOptions,
    ) -> JsonLdRecord | list[JsonLdRecord]:
        """Parse raw data into Linked Data."""
        raise NotImplementedError()



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
    JsonLdContext,
    JsonLdInput,
    JsonLdRecord,
)


class FlattenOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options to flatten a YAML-LD document."""

    compact_arrays: bool = True
    """Compact arrays to single values when appropriate?"""


DEFAULT_FLATTEN_OPTIONS = FlattenOptions()   # type: ignore


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def flatten(
    document: JsonLdInput,
    ctx: JsonLdContext | None = None,
    options: FlattenOptions = DEFAULT_FLATTEN_OPTIONS,
) -> JsonLdRecord:
    """Flatten a document."""
    with except_json_ld_errors():
        return jsonld.flatten(
            input_=str(document),
            ctx=ctx,
            options=options.model_dump(by_alias=True),
        )



from yaml_ld.compact import compact
from yaml_ld.expand import expand
from yaml_ld.flatten import flatten  # noqa: WPS347
from yaml_ld.frame import frame
from yaml_ld.from_rdf import from_rdf  # noqa: WPS347
from yaml_ld.load_document import load_document  # noqa: WPS347
from yaml_ld.to_rdf import to_rdf  # noqa: WPS347

__all__ = ['expand', 'compact', 'to_rdf', 'from_rdf', 'flatten', 'frame', 'load_document']   # noqa: WPS410



from typing import Literal

from typing_extensions import TypedDict

TermType = Literal['IRI', 'blank node', 'literal']


class IRITerm(TypedDict):
    """RDF Term."""

    type: Literal['IRI']
    value: str   # noqa: WPS110


class BlankTerm(TypedDict):
    """RDF Term."""

    type: Literal['blank node']
    value: str   # noqa: WPS110


class LiteralTerm(TypedDict):
    """RDF Term."""

    type: Literal['literal']
    value: str   # noqa: WPS110
    datatype: str | None


Term = IRITerm | BlankTerm | LiteralTerm


class Triple(TypedDict):
    """RDF Triple."""

    subject: Term
    predicate: Term
    object: Term


Graph = list[Triple]
Dataset = dict[str, Graph]



from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Sequence

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel
from typing_extensions import TypedDict
from urlpath import URL

JsonLdRecord = dict[str, Any]  # type: ignore


class RemoteDocument(TypedDict):
    contentType: str
    contextUrl: str
    document: Any
    documentUrl: str
    profile: str


JsonLdInput = (
    JsonLdRecord | Sequence[JsonLdRecord] | str | Path | URL | RemoteDocument
)
"""
Input for `expand()`, `compact()` and other functions.

[Specification](https://w3c.github.io/json-ld-api/#dom-jsonldrecord)
"""

JsonLdContext = JsonLdRecord | list[JsonLdRecord | str] | str
"""
The `JsonLdContext` interface is used to refer to a value that may be a
`JsonLdRecord`, a sequence of `JsonLdRecord`-s, or a string representing an IRI,
which can be dereferenced to retrieve a valid JSON document.

[Specification](https://w3c.github.io/json-ld-api/#dom-jsonldcontext)
"""


ExtractAllScripts = Annotated[
    bool,
    'Extract all JSON-LD script elements, as opposed to only the first one.',
]


class ExtractAllScriptsOptions(BaseModel):    # type: ignore
    """Options flag to extract all scripts or not."""

    extract_all_scripts: bool = False
    """
    True to extract all YAML-LD script elements from HTML, False to extract just
    the first.
    """


class ExpandContextOptions(BaseModel):    # type: ignore
    """Options flag to extract all scripts or not."""

    expand_context: JsonLdRecord | str | Path | URL | None = None
    """A context to expand with."""

    model_config = ConfigDict(arbitrary_types_allowed=True)


def _default_document_loader():
    from yaml_ld.document_loaders.default import (  # noqa: WPS433
        DEFAULT_DOCUMENT_LOADER,
    )
    return DEFAULT_DOCUMENT_LOADER


class BaseOptions(BaseModel):   # type: ignore
    """Base options shared by all YAML-LD API methods."""

    base: str | Path | None = None
    """The base IRI to use."""

    document_loader: Annotated[  # type: ignore
        Any,
        Field(default_factory=_default_document_loader),
    ]
    """The document loader."""

    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=to_camel,
        arbitrary_types_allowed=True,
    )


DEFAULT_VALIDATE_CALL_CONFIG = ConfigDict(arbitrary_types_allowed=True)



import datetime

import yaml


def xsd(term: str) -> str:
    return f'http://www.w3.org/2001/XMLSchema#{term}'


def i18n(term: str) -> str:
    return f'https://www.w3.org/ns/i18n#{term}'


def tag(term: str) -> str:
    return f'tag:yaml.org,2002:{term}'


class YAMLLDLoader(yaml.SafeLoader):
    pass

CONSTRUCTORS = {
    tag('timestamp'): YAMLLDLoader.construct_scalar,
    xsd('integer'): YAMLLDLoader.construct_yaml_int,
    xsd('decimal'): YAMLLDLoader.construct_yaml_float,
    xsd('double'): YAMLLDLoader.construct_scalar,
    xsd('boolean'): YAMLLDLoader.construct_yaml_bool,
    xsd('date'): YAMLLDLoader.construct_scalar,
    xsd('time'): YAMLLDLoader.construct_scalar,
    xsd('dateTime'): YAMLLDLoader.construct_scalar,
    i18n('en-US'): YAMLLDLoader.construct_yaml_str,
    i18n('en-US_ltr'): YAMLLDLoader.construct_yaml_str,
    i18n('_rtl'): YAMLLDLoader.construct_yaml_str,
}


for yaml_tag, constructor in CONSTRUCTORS.items():
    YAMLLDLoader.add_constructor(yaml_tag, constructor)   # type: ignore



from typing import Annotated

from pydantic import validate_call
from pyld import jsonld

from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
    JsonLdContext,
    JsonLdInput,
    JsonLdRecord,
)


class CompactOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options to compact a YAML-LD document."""

    compact_arrays: bool = True
    """Compact arrays to single values when appropriate?"""

    graph: bool = False
    """True to always output a top-level graph."""

    skip_expansion: bool = False
    """True to skip the expansion process, False to include it."""


DEFAULT_COMPACT_OPTIONS = CompactOptions()   # type: ignore


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def compact(  # noqa: WPS211
    document: JsonLdInput,
    ctx: Annotated[JsonLdContext | None, 'Context to compact with.'],
    options: CompactOptions = DEFAULT_COMPACT_OPTIONS,
) -> JsonLdRecord | list[JsonLdRecord]:
    """Compact a JSON-LD document."""
    with except_json_ld_errors():
        return jsonld.compact(
            input_=str(document),
            ctx=ctx,
            options=options.model_dump(by_alias=True),
        )



from pathlib import Path

from urlpath import URL


def as_url_or_path(raw: str) -> URL | Path:
    """Interpret a raw string as a URL or a local disk path."""
    if (url := URL(raw)).scheme:
        return url

    return Path(raw)



from dataclasses import dataclass
from pathlib import Path

from documented import DocumentedError

from yaml_ld.models import JsonLdRecord


class YAMLLDError(DocumentedError):
    """An error happened while processing YAML-LD data."""


@dataclass
class PyLDError(YAMLLDError):   # type: ignore
    """{self.message}"""   # noqa: D400

    message: str
    code: str


@dataclass
class DocumentIsScalar(YAMLLDError):   # type: ignore
    """
    Document content MUST be sequence or mapping.

    Instead, `{self.document_type_name}` found.
    """

    document: JsonLdRecord
    code: str = 'loading document failed'

    @property
    def document_type_name(self):
        return type(self.document).__name__


@dataclass
class LoadingDocumentFailed(YAMLLDError):   # type: ignore
    """
    Document is not a valid YAML.

    Path: {self.path}
    """

    path: str
    code: str = 'loading document failed'


@dataclass
class NotFound(YAMLLDError):   # type: ignore
    """
    Document has not been found.

    Path: {self.path}
    """

    path: str
    code: str = 'loading document failed'


@dataclass
class InvalidScriptElement(YAMLLDError):   # type: ignore
    """HTML <script> element content is not valid YAML."""

    code: str = 'invalid script element'


@dataclass
class NoLinkedDataFoundInHTML(YAMLLDError):   # type: ignore
    """No Linked Data fragments found in an HTML document."""

    code: str = 'loading document failed'


@dataclass
class MappingKeyError(YAMLLDError):   # type: ignore
    """A mapping key MUST be a string."""

    code: str = 'mapping-key-error'


@dataclass
class InvalidJSONLiteral(YAMLLDError):   # type: ignore
    """A mapping key MUST be a string."""

    code: str = 'invalid JSON literal'


@dataclass
class InvalidEncoding(YAMLLDError):   # type: ignore
    """
    A YAML-LD document MUST be encoded in UTF-8.

    To ensure interoperability with [[JSON]].
    """

    code: str = 'invalid encoding'


@dataclass
class CycleDetected(YAMLLDError):   # type: ignore
    """A YAML-LD document MUST NOT contain cycles."""

    code: str = 'loading document failed'


@dataclass
class UndefinedAliasFound(YAMLLDError):   # type: ignore
    """An undefined alias found."""

    code: str = 'loading document failed'


@dataclass
class LoadingRemoteContextFailed(YAMLLDError):   # type: ignore
    """
    Failed to load the context.

    URL of the context: {self.context}
    Reason: {self.reason}
    """

    context: str
    reason: str

    code: str = 'loading remote context failed'



from dataclasses import dataclass
from pathlib import Path

from yaml_ld.document_loaders.base import (
    DocumentLoader,
    DocumentLoaderOptions,
    PyLDResponse,
)


@dataclass
class MockLoader(DocumentLoader):
    response: PyLDResponse

    def __call__(
        self,
        source: str | Path,
        options: DocumentLoaderOptions,
    ) -> PyLDResponse:
        return self.response



from pathlib import Path

from urlpath import URL

from yaml_ld.document_loaders import content_types
from yaml_ld.document_loaders.base import (
    DocumentLoader,
    DocumentLoaderOptions,
    PyLDResponse,
)
from yaml_ld.errors import LoadingDocumentFailed, NotFound


class LocalFileDocumentLoader(DocumentLoader):
    """Load documents from a local file system."""

    def __call__(
        self,
        source: str | Path,
        options: DocumentLoaderOptions,
    ) -> PyLDResponse:
        """Load documents from a local file system."""
        path = Path(URL(source).path)

        content_type = content_types.by_extension(path.suffix)
        if content_type is None:
            raise ValueError(f'What content type is extension {path.suffix}?')

        parser = content_types.parser_by_content_type(content_type)
        if parser is None:
            raise LoadingDocumentFailed(path=path)

        try:
            with path.open() as f:
                yaml_document = parser(f, str(source), options)
        except FileNotFoundError as file_not_found:
            raise NotFound(path) from file_not_found

        return {
            'document': yaml_document,
            'documentUrl': str(source),
            'contextUrl': None,
            'contentType': content_type,
        }



import functools

from yaml_ld.document_loaders.choice_by_scheme import (
    ChoiceBySchemeDocumentLoader,
)
from yaml_ld.document_loaders.http import HTTPDocumentLoader
from yaml_ld.document_loaders.local_file import LocalFileDocumentLoader

DEFAULT_DOCUMENT_LOADER = ChoiceBySchemeDocumentLoader(
    file=LocalFileDocumentLoader(),
    http=HTTPDocumentLoader(),
    https=HTTPDocumentLoader(),
)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable

import funcy
from documented import DocumentedError
from pydantic import validate_call
from urlpath import URL

from yaml_ld.document_loaders.base import (
    DocumentLoader,
    DocumentLoaderOptions,
    PyLDResponse,
)


@dataclass
class ProtocolNotFound(DocumentedError):  # type: ignore
    """
    Cannot choose the loader by URL protocol.

    * URL: `{self.url}`
    * Scheme: `{self.formatted_scheme}`
    * Available schemes: {self.formatted_schemes}
    """

    url: URL
    schemes: Iterable[str]

    @property
    def formatted_scheme(self):
        return self.url.scheme or '(empty string)'

    @property
    @funcy.joining(', ')
    def formatted_schemes(self):
        return self.schemes


class ChoiceBySchemeDocumentLoader(DocumentLoader):
    loaders: dict[str, DocumentLoader]

    def __init__(self, **loaders: DocumentLoader) -> None:
        self.loaders = loaders

    def __call__(
        self,
        source: str | Path,
        options: DocumentLoaderOptions,
    ) -> PyLDResponse:
        url = URL(source)

        try:
            loader = self.loaders[url.scheme or 'file']
        except KeyError:
            raise ProtocolNotFound(
                schemes=self.loaders.keys(),
                url=url,
            )

        return loader(source, options)



import contextlib
from json import JSONDecodeError
from pathlib import Path

from pydantic import validate_call
from pyld import jsonld
from urlpath import URL

from yaml_ld.errors import (
    CycleDetected,
    InvalidJSONLiteral,
    InvalidScriptElement,
    LoadingRemoteContextFailed,
    MappingKeyError,
    PyLDError,
)
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
    JsonLdInput,
    JsonLdRecord,
)


class ExpandOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options for `jsonld.expand()`."""


DEFAULT_EXPAND_OPTIONS = ExpandOptions()   # type: ignore


def except_json_ld_error(err: jsonld.JsonLdError):
    """Handle JsonLdError."""
    # We need to drill down; for instance, `to_rdf()` raises an error which
    # contains an actual error from `expand()` in its `.cause` field.
    err = err.cause or err

    if isinstance(err, JSONDecodeError):
        raise InvalidJSONLiteral() from err

    match err.code:
        case LoadingRemoteContextFailed.code:
            raise LoadingRemoteContextFailed(
                context=err.details['url'],
                reason=str(err.details['cause']),
            ) from err

        case _:
            raise PyLDError(
                message=str(err),
                code=err.code,
            ) from err


@contextlib.contextmanager
def except_json_ld_errors():
    """Convert pyld errors to typed YAML-LD exceptions."""
    try:
        yield
    except TypeError as err:
        raise MappingKeyError() from err
    except RecursionError as err:
        raise CycleDetected() from err
    except JSONDecodeError as err:
        raise InvalidScriptElement() from err
    except jsonld.JsonLdError as err:
        except_json_ld_error(err)


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def expand(   # noqa: C901, WPS211
    document: JsonLdInput,
    options: ExpandOptions = DEFAULT_EXPAND_OPTIONS,
) -> list[JsonLdRecord]:
    """Expand a YAML-LD document."""
    with except_json_ld_errors():
        return jsonld.expand(
            input_=str(document) if isinstance(document, (URL, Path)) else document,
            options=options.model_dump(
                by_alias=True,
                exclude_none=True,
            ),
        )



import io

import yaml
from yaml.composer import ComposerError
from yaml.constructor import ConstructorError
from yaml.parser import ParserError
from yaml.scanner import ScannerError

from yaml_ld.document_parsers.base import (
    BaseDocumentParser,
    DocumentLoaderOptions,
)
from yaml_ld.errors import (
    DocumentIsScalar,
    InvalidEncoding,
    InvalidScriptElement,
    LoadingDocumentFailed,
    MappingKeyError,
    UndefinedAliasFound,
)
from yaml_ld.loader import YAMLLDLoader
from yaml_ld.models import JsonLdRecord


def _ensure_not_scalar(document) -> JsonLdRecord | list[JsonLdRecord]:
    if not isinstance(document, (dict, list)):
        raise DocumentIsScalar(document)

    return document


class YAMLDocumentParser(BaseDocumentParser):
    """Parse YAML documents."""

    def __call__(   # noqa: WPS238, WPS231, WPS225, C901
        self,
        data_stream: io.TextIOBase,
        source: str,
        options: DocumentLoaderOptions,
    ) -> JsonLdRecord | list[JsonLdRecord]:
        """Parse YAML document stream into LD."""
        yaml_documents_stream = yaml.load_all(  # noqa: S506
            stream=data_stream,
            Loader=YAMLLDLoader,
        )

        try:   # noqa: WPS225
            return _ensure_not_scalar(
                self._yaml_document_from_stream(
                    stream=yaml_documents_stream,
                    extract_all_scripts=options.get('extractAllScripts', False),
                ),
            )

        except UnicodeDecodeError as unicode_decode_error:
            raise InvalidEncoding() from unicode_decode_error

        except ConstructorError as err:
            if err.problem == 'found unhashable key':
                raise MappingKeyError() from err

            raise

        except ScannerError as err:
            raise LoadingDocumentFailed(path=source) from err

        except ComposerError as err:
            raise UndefinedAliasFound() from err

        except ParserError as err:
            raise InvalidScriptElement() from err

    def _yaml_document_from_stream(self, stream, extract_all_scripts: bool):
        if extract_all_scripts:
            return list(stream)

        try:
            return next(stream)
        except StopIteration as stop_iteration:
            raise LoadingDocumentFailed(path='') from stop_iteration



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import BaseOptions, JsonLdRecord


class FromRDFOptions(BaseOptions):   # type: ignore
    """Options for converting RDF to YAML-LD."""

    format: str = 'application/n-quads'
    """The format if input is a string: 'application/n-quads' for N-Quads."""

    use_rdf_type: bool = False
    """True to use `rdf:type`, False to use `@type`."""

    use_native_types: bool = False
    """Convert XSD types into native types (boolean, integer, double)?"""


@validate_call(config=dict(arbitrary_types_allowed=True))
def from_rdf(
    dataset: str,
    options: FromRDFOptions = FromRDFOptions(),  # type: ignore
) -> JsonLdRecord:
    """Convert a RDF dataset to a YAML-LD document."""
    with except_json_ld_errors():
        return jsonld.from_rdf(dataset, options.model_dump(by_alias=True))



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
    JsonLdInput,
    JsonLdRecord,
)


class FrameOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options for YAML-LD framing."""

    embed: bool = True
    """
    Sets the value object embed flag used in the Framing Algorithm. A boolean
    value of true sets the flag to @once, while a value of false sets the flag
    to @never.
    """

    explicit: bool = False
    """Default `@explicit` flag."""

    omit_default: bool = False
    """Default `@omit_default` flag."""

    prune_blank_node_identifiers: bool = True
    """Remove unnecessary blank node identifiers."""

    require_all: bool = False
    """Default `@require_all` flag."""

    frame_default: bool = False
    """Instead of framing a merged graph, frame only the default graph."""


@validate_call(config=dict(arbitrary_types_allowed=True))
def frame(
    document: JsonLdInput,
    frame: JsonLdRecord,
    options: FrameOptions = FrameOptions(),   # type: ignore
) -> JsonLdRecord:
    """Frame a YAML-LD document."""
    with except_json_ld_errors():
        return jsonld.frame(
            input_=str(document),
            frame=frame,
            options=options.model_dump(by_alias=True),
        )
